""" this module calculate evaluation metrics for pyspark mlib """

from pyspark.sql import DataFrame as SparkDataFrame
import pyspark.pandas as ps
from pyspark.ml.evaluation import (
    BinaryClassificationEvaluator,
    MulticlassClassificationEvaluator,
)


def evaluator(
    df: SparkDataFrame,
    model_name: str,
    target_column: str,
    data_set_type: str = "training",
):
    # AUC
    bc = BinaryClassificationEvaluator(
        rawPredictionCol="rawPrediction",
        labelCol=target_column,
        metricName="areaUnderROC",
    )
    auc = bc.evaluate(df)
    # Accuracy
    mc = MulticlassClassificationEvaluator(
        labelCol="prediction", predictionCol=target_column
    )
    accuracy = mc.evaluate(df, {mc.metricName: "accuracy"})
    # precision
    precision = mc.evaluate(df, {mc.metricName: "weightedPrecision"})
    # recall
    recall = mc.evaluate(df, {mc.metricName: "weightedRecall"})
    # f1
    f1 = mc.evaluate(df, {mc.metricName: "f1"})

    results = ps.DataFrame(
        dict(
            model=[model_name],
            type=[data_set_type],
            auc=[auc],
            accuracy=[accuracy],
            weighted_precision=[precision],
            weighted_recall=[recall],
            f1=[f1],
        )
    )

    return results



""" This applies Regression models available in pyspark"""
from typing import List


from pyspark.sql import DataFrame as SparkDataFrame
import pyspark.pandas as ps


from SparkAutoML.ml_module.preprocessing_module.preprocess_file import Preprocessor
from SparkAutoML.utils.models_dict_file import model_dict_regression
from SparkAutoML.utils.Regression_evaluator_file import evaluator


class Regressor(Preprocessor):
    def create_model(self, model_name: str, **args) -> None:
        print(f"Training started for {model_name} ....")
        # instantiate the model
        self.model_name = model_name
        self.model = model_dict_regression[model_name]
        self.model = self.model(
            featuresCol="features", labelCol=self.target_feature, **args
        )

        # run the preprocessing pipeline
        self.run_pipeline()
        pipe = self.pipeline.getStages()
        pipe.append(self.model)
        self.pipeline.setStages(pipe)

        # now do fit transform
        self.fit_transform()
        # fit on hold out data set
        self.transform()
        print(f"Training successfully ended for {model_name} ....")

    def evaluate_model(self, evaluate_on: str = "train") -> None:
        """evaluation on train or holdout data set"""
        if evaluate_on == "train":

            results = evaluator(
                self.train_data_transformed,
                self.model_name,
                self.target_feature,
                data_set_type=evaluate_on,
            )
            self.evaluation_results_training = results
        else:
            results = evaluator(
                self.hold_out_data_transformed,
                self.model_name,
                self.target_feature,
                data_set_type=evaluate_on,
            )
            self.evaluation_results_hold_out = results

        return results

    def compare_models(self, sort_by: str = "r2"):
        """use available models in mlib and evaluate them on holdout dataset"""

        results = ps.DataFrame()
        self.compare_model_dict = {}
        # create a placeholder
        for model in model_dict_regression.keys():
            # try to train all models
            try:
                self.create_model(model)
                eval = self.evaluate_model(evaluate_on="holdout")
                results = results.append(eval)
                self.compare_model_dict[model] = self.fitted_pipeline
            except:
                pass

        # sort results by user choice
        results = results.reset_index(drop=True).sort_values(
            by=sort_by, ascending=False
        )

        # keep the top performing pipeline
        self.fitted_pipeline = self.compare_model_dict[results.iloc[0, 0]]
        self.evaluation_results_hold_out = results

        return results

    def predict_model(self, df: SparkDataFrame) -> SparkDataFrame:
        """make prediction on future dataset (data without label) """

        # keep the top performing pipeline
        results = self.fitted_pipeline.transform(df)

        return results



import logging

logging.basicConfig(level=logging.DEBUG)
logging.basicConfig(format="%(asctime)s - %(message)s", datefmt="%d-%b-%y %H:%M:%S")
# info level logger


def info_log(message: str):
    return logging.info(message)




# make a dictionary of available models in pyspark

from pyspark.ml.classification import (
    LogisticRegression,
    GBTClassifier,
    RandomForestClassifier,
    DecisionTreeClassifier,
    MultilayerPerceptronClassifier,
    LinearSVC,
    NaiveBayes,
    FMClassifier,
)

from pyspark.ml.regression import (
    LinearRegression,
    GeneralizedLinearRegression,
    DecisionTreeRegressor,
    RandomForestRegressor,
    GBTRegressor,
    AFTSurvivalRegression,
    IsotonicRegression,
    FMRegressor,
)

model_dict_classifier = {
    "lr": LogisticRegression,
    "rfc": RandomForestClassifier,
    "gbc": GBTClassifier,
    "dtc": DecisionTreeClassifier,
    "mlpc": MultilayerPerceptronClassifier,
    "svc": LinearSVC,
    "nbc": NaiveBayes,
    "fmc": FMClassifier,
}


model_dict_regression = {
    "lr": LinearRegression,
    "glr": GeneralizedLinearRegression,
    "dtr": DecisionTreeRegressor,
    "rfr": RandomForestRegressor,
    "gbr": GBTRegressor,
    "sr": AFTSurvivalRegression,
    "isor": IsotonicRegression,
    "fmr": FMRegressor,
}



""" this module calculate evaluation metrics for pyspark mlib """


from pyspark.sql import DataFrame as SparkDataFrame
import pyspark.pandas as ps
from pyspark.ml.evaluation import RegressionEvaluator


def evaluator(
    df: SparkDataFrame,
    model_name: str,
    target_column: str,
    data_set_type: str = "training",
):
    # r2
    r2_evaluator = RegressionEvaluator(
        labelCol=target_column, predictionCol="prediction", metricName="r2"
    )
    r2 = r2_evaluator.evaluate(df)

    # rmse
    rmse_evaluator = RegressionEvaluator(
        labelCol=target_column, predictionCol="prediction", metricName="rmse"
    )
    rmse = rmse_evaluator.evaluate(df)

    # mse
    mse_evaluator = RegressionEvaluator(
        labelCol=target_column, predictionCol="prediction", metricName="mse"
    )
    mse = mse_evaluator.evaluate(df)

    # mae
    mae_evaluator = RegressionEvaluator(
        labelCol=target_column, predictionCol="prediction", metricName="mae"
    )
    mae = mae_evaluator.evaluate(df)

    # result holder
    results = ps.DataFrame(
        dict(
            model=[model_name],
            type=[data_set_type],
            r2=[round(r2, 4)],
            rmse=[round(rmse, 4)],
            mse=[round(mse, 4)],
            mae=[round(mae, 4)],
        )
    )

    return results



from typing import List
from pyspark.sql import DataFrame as SparkDataFrame

from pyspark.ml.feature import (
    VectorAssembler,
    VectorIndexer,
    OneHotEncoder,
    StringIndexer,
    Imputer,
)
from pyspark.ml import Pipeline
from pyspark.sql.types import DoubleType


class Preprocessor:
    """
  Preprocess to prepare data for ML tasks
  """

    def __init__(
        self,
        training_data: SparkDataFrame,
        hold_out_data: SparkDataFrame,
        target_feature: "str",
        numeric_features: List[str],
        categorical_features: List[str],
        impute_missing_values: bool = False,
        missing_value_strategy: str = "mean",
    ) -> None:

        self.train_data = training_data
        self.hold_out_data = hold_out_data
        self.target_feature = target_feature
        self.numeric_features = numeric_features
        self.categorical_features = categorical_features
        self.impute_missing_values = impute_missing_values
        self.missing_value_strategy = missing_value_strategy

    def _column_name_generator(self, input_cols: list, suffix: str) -> list:
        """This function adds suffix to the list of columns"""
        new_cols = [column + str(suffix) for column in input_cols]
        return new_cols

    def _cast_double_type(self, df: SparkDataFrame, column: str) -> SparkDataFrame:
        df = df.withColumn(column, df[column].cast(DoubleType()))
        return df

    # ====================== Transformers ===================================

    def _indexer(self, input_cols: list, suffix: str = "_index") -> None:
        """spark's indexer transformer"""
        self._indexed_cols = self._column_name_generator(input_cols, suffix)
        self.indexers = StringIndexer(
            inputCols=input_cols, outputCols=self._indexed_cols
        )
        return None

    def _encoder(self, input_cols: list, suffix: str = "_onehot") -> None:
        """spark's one hot encoder transformer"""
        self._encoded_cols = self._column_name_generator(input_cols, suffix)
        self.encoders = OneHotEncoder(
            inputCols=input_cols, outputCols=self._encoded_cols, dropLast=False
        )
        return None

    def _imputer(
        self, input_cols: list, suffix: str = "_imputed", strategy: str = "mean"
    ) -> None:
        self._imputed_cols = self._column_name_generator(input_cols, suffix)
        self.imputers = Imputer(
            inputCols=input_cols,
            outputCols=self._imputed_cols,
            strategy=self.missing_value_strategy,
        )
        return None

    def _vector_assembler(
        self, input_cols: list, output_cols: str = "features"
    ) -> None:
        """spark's Vector Assembler transformer"""
        self.vector = VectorAssembler(inputCols=input_cols, outputCol="features")
        return None

    def _pipeline(self, stages: list) -> None:
        """spark's Pipeline class """
        self.pipeline = Pipeline(stages=stages)
        return None

    # ===============================Pipelines==========================================
    def _pipeline_numeric(self) -> None:
        """->Vector Assembler"""
        self._vector_assembler(input_cols=self.numeric_features, output_cols="features")
        # pipeline
        self._pipeline([self.vector])
        return None

    def _pipeline_numeric_imputer(self) -> None:
        """->imputer -> Vector Assembler"""
        self._imputer(input_cols=self.numeric_features)
        self._vector_assembler(input_cols=self.numeric_features, output_cols="features")
        # pipeline
        self._pipeline([self.imputers, self.vector])

    def _pipeline_categorical(self) -> None:
        # transformers
        self._indexer(input_cols=self.categorical_features)
        self._encoder(input_cols=self._indexed_cols)
        self._vector_assembler(input_cols=self._encoded_cols, output_cols="features")
        # pipeline
        self._pipeline([self.indexers, self.encoders, self.vector])

    def _pipeline_mixed(self) -> None:
        """indexer -> Categorical-> Vector Assembler"""
        # transformers
        self._indexer(input_cols=self.categorical_features)
        self._encoder(input_cols=self._indexed_cols)
        self._vector_assembler(
            input_cols=self.numeric_features + self._encoded_cols,
            output_cols="features",
        )
        # pipeline
        self._pipeline([self.indexers, self.encoders, self.vector])
        return None

    def _pipeline_mixed_imputer(self) -> None:
        """indexer -> Categorical-> Vector Assembler"""
        # transformers
        self._indexer(input_cols=self.categorical_features)
        self._encoder(input_cols=self._indexed_cols)
        self._imputer(input_cols=self.numeric_features)
        self._vector_assembler(
            input_cols=self._imputed_cols + self._encoded_cols, output_cols="features"
        )
        # pipeline
        self._pipeline([self.indexers, self.encoders, self.imputers, self.vector])
        return None

    # ====================================Executions=================================================
    def run_pipeline(self):
        """ This is will run pipeline accoding to user selections"""

        # make sure target column is of double type
        self.train_data = self._cast_double_type(
            df=self.train_data, column=self.target_feature
        )
        self.hold_out_data = self._cast_double_type(
            df=self.hold_out_data, column=self.target_feature
        )

        # numeric only pipeline
        if (
            self.categorical_features is None
            and self.numeric_features is not None
            and self.impute_missing_values == False
        ):
            self._pipeline_numeric()

        # numeric + imputation
        if (
            self.categorical_features is None
            and self.numeric_features is not None
            and self.impute_missing_values == True
        ):
            self._pipeline_numeric_imputer()

        # categorical only
        if (
            self.categorical_features is not None
            and self.numeric_features is None
            and self.impute_missing_values == False
        ):
            self._pipeline_categorical()

        # mixed (numeric + categorical) pipeline
        if (
            self.categorical_features is not None
            and self.numeric_features is not None
            and self.impute_missing_values == False
        ):
            self._pipeline_mixed()

        # mixed + imputer  pipeline
        if (
            self.categorical_features is not None
            and self.numeric_features is not None
            and self.impute_missing_values == True
        ):
            self._pipeline_mixed_imputer()

        return None

    def fit_transform(self) -> None:
        """fit on the preprocessing file. Make sure to execute the run_pipeline method first"""
        self.fitted_pipeline = self.pipeline.fit(self.train_data)
        self.train_data_transformed = self.fitted_pipeline.transform(self.train_data)

        return None

    def transform(self) -> None:
        """transform hold out data"""
        self.hold_out_data_transformed = self.fitted_pipeline.transform(
            self.hold_out_data
        )

        return None




import pytest
import pyspark.pandas as ps

from SparkAutoML.ml_module.BinaryClassifier_module.classifier_file import BClassifier


@pytest.fixture
def sample_data_train():
    df = ps.DataFrame(
        dict(col_a=[1, 2, 3, 4], col_b=[0.24, 0.56, 0.12, 0.39], label=[1, 0, 0, 1])
    ).to_spark()
    return df


@pytest.fixture
def sample_data_holdout():
    df = ps.DataFrame(
        dict(col_a=[5, 6, 8], col_b=[0.30, 0.80, 0.32,], label=[1, 0, 0,])
    ).to_spark()
    return df


@pytest.fixture
def sample_data_unseen():
    df = ps.DataFrame(dict(col_a=[5, 6, 8], col_b=[0.30, 0.80, 0.32,],)).to_spark()
    return df


def test_create_model(sample_data_train, sample_data_holdout, sample_data_unseen):

    bc = BClassifier(
        training_data=sample_data_train,
        hold_out_data=sample_data_holdout,
        target_feature="label",
        numeric_features=["col_a", "col_b"],
        categorical_features=None,
    )

    # create model
    bc.create_model("lr")

    # make prediction
    pred = bc.predict_model(sample_data_unseen)

    # make sure that transformed data set has 'features' column
    assert "prediction" in pred.columns


def test_compare_model(sample_data_train, sample_data_holdout, sample_data_unseen):

    bc = BClassifier(
        training_data=sample_data_train,
        hold_out_data=sample_data_holdout,
        target_feature="label",
        numeric_features=["col_a", "col_b"],
        categorical_features=None,
    )

    # create model
    bc.compare_models()

    # make prediction
    pred = bc.predict_model(sample_data_unseen)

    # make sure that transformed data set has 'features' column
    assert "prediction" in pred.columns




from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))


def readme():
    with open("README.md", encoding="utf8") as f:
        README = f.read()
    return README

VERSION = "0.0.8"
DESCRIPTION = "For an easy implementation of spark's ml  "


# Setting up
setup(
    name="SparkAutoML",
    version=VERSION,
    author="Fahad Akbar",
    author_email="<fahadakbar@gmail.com>",
    description=DESCRIPTION,
    long_description=readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=["pandas", "numpy", "pyspark >= 3.2.1", "pytest"],
    keywords=["python", "spark", "machine learning",],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)
